#include "agent.hpp"

// The source register is taken as srcVal % REGISTERS so srcVal should be
// a multiple of REGISTERS so that there is no bias towards any register.
// For example, if REGISTERS is 4 and there srcVal can take values 0 to
// 4, then register 0 is expected to occur more frequently. The destination
// register is simply the bits in the dst field converted to ulong.

const instruction Agent::_modeMask(0x1);   // 0 0000 0000 0001
const instruction Agent::_opMask(0xE);     // 0 0000 0000 1110
const instruction Agent::_dstMask(0x70);   // 0 0000 0111 0000
const instruction Agent::_srcMask(0x1F80); // 1 1111 1000 0000

const instruction Agent::_mode0(0x0); // Rx <- op Rx Ry
const instruction Agent::_mode1(0x1); // Rx <- op Rx Iy

const instruction Agent::_opSum(0x0);
const instruction Agent::_opDiff(0x1);
const instruction Agent::_opProd(0x2);
const instruction Agent::_opDiv(0x3);
const instruction Agent::_opMod(0x4);
const instruction Agent::_opCos(0x5);
const instruction Agent::_opExp(0x6);
const instruction Agent::_opLog(0x7);

const short Agent::_modeShift = 0;
const short Agent::_opShift = Agent::_modeMask.count();
const short Agent::_dstShift = Agent::_modeMask.count() + Agent::_opMask.count();
const short Agent::_srcShift = Agent::_modeMask.count() + Agent::_opMask.count() + Agent::_dstMask.count();

unsigned long Agent::_count = 0;

ostream & operator<<(ostream & os, const Agent &agt)
{
  os << "[" << agt._id;
  //  os << ", " << agt._W;
  os << ", " << agt._old;
  os << ", " << agt._action;
  os << "]";

  return os; 
}

Agent::Agent(Args &args, long epoch, long action)
{
  _dim = args.getLong("numFeatures");
  _id = Agent::_count++;
  _action = action;
  _W = 0;
  _birthEpoch = epoch;
  _numWon = 0;
  _sumDue = 0;
  _sumReward = 0;
  _numProfitWon = 0;
  _sumProfitDue = 0;
  _sumProfitReward = 0;
  _profitRate = 1;

  _parentId = _id;
  _parentPtr = this;
  
  instruction *instr;
  int progSize = args.getLong("minIndSize") + ((int) (drand48() * (args.getLong("maxIndSize") - args.getLong("minIndSize"))));
  
  for(int i = 0; i < progSize; i++)
    {
      instr = new instruction();
      
      for(int j = 0; j < instr->size(); j++)
	if(drand48() < 0.5) instr->flip(j);
      
      _bid.push_back(instr);
    }
}

Agent::Agent(Args &args, long epoch, Agent &parent, vector < int > &actions, double pActionMutate)
{
  _dim = args.getLong("numFeatures");
  _id = Agent::_count++;
  _action = parent._action;
  _W = 0;
  _birthEpoch = epoch;
  _numWon = 0;
  _sumDue = 0;
  _sumReward = 0;
  _numProfitWon = 0;
  _sumProfitDue = 0;
  _sumProfitReward = 0;
  _profitRate = 1;

  _parentId = parent.id();
  _parentPtr = &parent;

  instruction *tmp;

  vector < instruction * >::iterator iter;

  for(iter = parent._bid.begin(); iter != parent._bid.end(); iter++)
    {
      _bid.push_back(new instruction(**iter));
    }
 
  // remove random instruction
  if(_bid.size() > args.getLong("minIndSize") && drand48() < args.getDouble("pBidDelete"))
    {
      int i = (int) (drand48() * _bid.size());
      _bid.erase(_bid.begin() + i);
    }
  
  // insert random instruction
  if(_bid.size() < args.getLong("maxIndSize") && drand48() < args.getDouble("pBidAdd"))
    {
      instruction *instr = new instruction();
      
      for(int j = 0; j < instr->size(); j++)
        if(drand48() < 0.5) instr->flip(j);
      
      int i = (int) (drand48() * _bid.size());

      _bid.insert(_bid.begin() + i, instr);
    }

  // flip single bit of random instruction
  if(drand48() < args.getDouble("pBidMutate"))
    {
      int i = (int) (drand48() * _bid.size());
      int j = (int) (drand48() * _bid[0]->size());
      
      _bid[i]->flip(j);
    }

  // swap positions of two instructions
  if(drand48() < args.getDouble("pBidSwap"))
    {
      int i = (int) (drand48() * _bid.size());
      int j = (int) (drand48() * _bid.size());
      
      tmp = _bid[i];
      _bid[i] = _bid[j];
      _bid[j] = tmp;
    }
    
  // mutate
  if(drand48() < pActionMutate)
    {
      _action = actions[(int) (drand48() * actions.size())];
    }  
}

Agent::~Agent()
{
  for(int i = 0; i < _bid.size(); i++)
    delete _bid[i];
}

double Agent::bid(double *feature, double *REG)
{
  memset(REG, 0, REGISTERS * sizeof(double)); // Zero registers.
  return 1 / (1 + exp(-run(_bid, feature, REG)));
}

double Agent::run(vector < instruction * > &program, double *feature, double *REG)
{  
  vector < instruction * >::iterator iter;

  instruction mode;
  instruction op;
  
  int dstReg;
  double srcVal;

#ifdef CCM_SHOWPROG
  cout << *this << endl << "FEATURE";
  for(int i = 0; i < _dim; i++)
    cout << " " << feature[i];
  cout << endl << "REG";
  for(int i = 0; i < REGISTERS; i++)
    cout << " " << REG[i];
  cout << endl;
#endif
  
  for(iter = program.begin(); iter != program.end(); iter++)
    {
      mode = (**iter & _modeMask) >> _modeShift; 
      op = (**iter & _opMask) >> _opShift;

      dstReg = ((**iter & _dstMask) >> _dstShift).to_ulong();

#ifdef CCM_SHOWPROG
      cout << **iter << " R[" << dstReg << "] <- R[" << dstReg << "] ";
#endif

      if(mode == _mode0) // Rx <- op Rx Ry
	{
	  srcVal = REG[((**iter & _srcMask) >> _srcShift).to_ulong() % REGISTERS];
#ifdef CCM_SHOWPROG
	  cout << "R[" << ((**iter & _srcMask) >> _srcShift).to_ulong() % REGISTERS << "] ";
#endif
	}
      else // Rx <- op Rx Iy
	{
	  srcVal = feature[((**iter & _srcMask) >> _srcShift).to_ulong() % _dim];
#ifdef CCM_SHOWPROG
	  cout << "I[" <<  ((**iter & _srcMask) >> _srcShift).to_ulong() % _dim << "] ";
#endif
	}

      if(op == _opSum)
	{
	  REG[dstReg] = REG[dstReg] + srcVal;
#ifdef CCM_SHOWPROG
	  cout << "sum ";
#endif
	}
      else if(op == _opDiff)
	{
	  REG[dstReg] = REG[dstReg] - srcVal;
#ifdef CCM_SHOWPROG
	  cout << "diff ";
#endif
	}
      else if(op == _opProd)
	{
	  REG[dstReg] = REG[dstReg] * srcVal;
#ifdef CCM_SHOWPROG
	  cout << "prod ";
#endif
	}
      else if(op == _opDiv)
	{
	  REG[dstReg] = REG[dstReg] / srcVal;
#ifdef CCM_SHOWPROG
	  cout << "div ";
#endif
	}
      else if(op == _opMod)
	{
	  REG[dstReg] = fmod(REG[dstReg], srcVal);
#ifdef CCM_SHOWPROG
	  cout << "mod ";
#endif
	}
      else if(op == _opCos)
	{
	  REG[dstReg] = cos(srcVal);
#ifdef CCM_SHOWPROG
	  cout << "cos ";
#endif
	}
      else if(op == _opExp)
	{
	  REG[dstReg] = exp(srcVal);
#ifdef CCM_SHOWPROG
	  cout << "exp ";
#endif
	}
      else if(op == _opLog)
	{
	  REG[dstReg] = log(fabs(srcVal));
#ifdef CCM_SHOWPROG
	  cout << "log ";
#endif
	}
      else
	{
	  die(__FILE__, __FUNCTION__, __LINE__, "bad operation");
	}
      
      if(isfinite(REG[dstReg]) == 0)
	REG[dstReg] = 0;
      
#ifdef CCM_SHOWPROG
      cout << "REG";
      for(int i = 0; i < REGISTERS; i++)
	cout << " " << REG[i];
      cout << endl;
#endif
    }
  
  return REG[0];
}

// bool Agent::outbids(Agent &comp)
// {
//   if(isEqual(_dummy, comp._dummy))
//     return winsTie(comp);
  
//   return _dummy > comp._dummy;
// }

bool Agent::winsTie(Agent &comp)
{
  //  if(isEqual(_W, comp._W))
  //    {
      if(_bid.size() == comp._bid.size())
	{
	  return _id > comp._id;
	}
      
      return _bid.size() < comp._bid.size();
      //}
  
//      return _W > comp._W;
}

void Agent::recordWin(double due, double reward)
{
  double BETA = 0.9;
  
  _numWon++;
  _sumDue += due;
  _sumReward += reward;

  _profitRate = BETA * _profitRate;
  
  if(isEqual(due, reward) || due < reward) // no loss
    {
      _numProfitWon++;
      _sumProfitDue += due;
      _sumProfitReward += reward;
      
      _profitRate += (1 - BETA) * 1;
      
      //      _profitRate = BETA * _profitRate + (1 - BETA) * 1;
    }
  
  //  else _profitRate = BETA * _profitRate;
}
