#include <iostream>
#include <cfloat>
#include <list>

#include <sys/resource.h>

#include "args.hpp"
#include "dataset.hpp"
#include "misc.hpp"
#include "agent.hpp"
#include "pop.hpp"
#include "stats.hpp"
#include "point.hpp"

// with SWB the learner can always bid max and not be penalized for getting
// something wrong and always be in 1st front

void train(Args &);

void initPop(int, Args &, Pop &, vector < int > &, int); // initialize random learner population
void initPop(int, Args &, Pop &, vector < int > &, int, set < Point * > &, DataSet &); // initialize random learner population using points
void initPoints(int, long, set < Point * > &, map < int, int > &, DataSet &); // initialize random point population

void genPop(Pop &, Args &, long, vector < int > &, int);
void genPoints(long, set < Point * > &, DataSet &);

Agent * auction(Pop &, DataSet &, Point &);
void setOutcome(Point &, Pop &, int, Agent *);

void collectOutcomes(Pop &, set < Point * > &, map < long, vector < double > * > &, vector < int > &);
void collectDistincts(set < Point * > &, Pop &, map < long, vector < int > * > &);
void normalizeOutcomes(map < long, vector < double > * > &, vector < int > &, int);

bool dominated(vector < double > &, vector < double > &);
bool dominated(vector < double > &, vector < double > &, double);
int dominatedOrEqual(vector < double > &, vector < double > &, double);
bool dominated(vector < int > &, vector < int > &);

void selectPop(Pop &, map < long, vector < double > * > &, long &);
void selectPoints(set < Point * > &, map < long, vector < int > * > &, map < int, int > &, long &);

void reduce(Pop &, map < long, vector < double > * > &);

void test(Pop &, DataSet &, DataSet &, Args &);

// int fnds(map < long, vector < double > * > &, map < long, int > &); // fast non-dominated sort

bool getusage(double &, double &);

int main(int argc, char **argv)
{
  if(argc != 2)
    die(__FILE__, __FUNCTION__, __LINE__, "parameter file not specified");
  
  Args args(argv[1]);
  cout << args << endl;
  train(args);
  
//   struct rusage res;
//   if(getrusage(RUSAGE_SELF, &res) == -1)
//     {
//       cout << "RUSAGE ERROR";
//     }
//   else
//     {
//       double sys = (double) res.ru_stime.tv_sec + res.ru_stime.tv_usec * 1e-6;
//       double usr = (double) res.ru_utime.tv_sec + res.ru_utime.tv_usec * 1e-6;
      
//       cout << "RUSAGE SYS " << sys << " USR " << usr << " SUM " << sys + usr << endl;
//     }
  
  return 0; 
}


bool getusage(double &sys, double &usr)
{
  struct rusage res;
  if(getrusage(RUSAGE_SELF, &res) == -1)
    {
      cout << "RUSAGE ERROR";
      return false;
    }
  else
    {
      sys = (double) res.ru_stime.tv_sec + res.ru_stime.tv_usec * 1e-6;
      usr = (double) res.ru_utime.tv_sec + res.ru_utime.tv_usec * 1e-6;
      return true;
    }
}

void train(Args &args)
{
  cout.precision(3);
  cout << fixed;

  srand48(args.getLong("seed"));
  
  DataSet trainData(args.getString("trainDataFile"), args.getLong("numFeatures"), args.getLong("trainNumPatterns"));
  cout << endl;
  //  Stats trainStats(&trainData);
  
  Pop solution; // union of all populations trained so far

  double sys1, usr1;
  if(!getusage(sys1, usr1))
    die(__FILE__, __FUNCTION__, __LINE__, "cannot get rusage");

  // train each subpopulation separately
  for(int action = 0; action < trainData.numCategories(); action++)
    { 
      cout << PROGR << " START TRAINING FOR ACTION " << trainData._categories[action] << endl;
      
      long epoch;
      
      Pop pop; // population being trained
      vector < Agent * >::iterator iter;
      
      set < Point * > PP; // point population
      set < Point * >::iterator pIter; // point populatin iterator
      
      map < long, vector < double > * > outcomes; // (learner id) -> (outcome vector)
      vector < int > labels; // label of the tests corresponding to the outcomes in vector 'outcomes'
      map < long, vector < double > * > :: iterator oIter;
      
      map < long, vector < int > * > distincts; // (point id) -> (distinction vector)
      map < long, vector < int > * > :: iterator dIter;
      
      map < int, int > pCount; // count patterns of each class
      map < int, int > :: iterator cIter;
      
      int index;

      long lDom = 0; // how many times a learner in the archive was dominated
      long pDom = 0; // how many times a point in the archive was dominated
      
      for(int i = 0; i < trainData._categories.size(); i++)
	pCount.insert(map < int, int > :: value_type(trainData._categories[i], 0));
      
      // initialize populations
      initPoints(args.getLong("numTests"), trainData.size() - 1, PP, pCount, trainData);
      initPop(args.getLong("numAgents"), args, pop, trainData._categories, action, PP, trainData);
      //initPop(args.getLong("numAgents"), args, pop, trainData._categories, action);
      
      // start epoch
      for(epoch = 0; epoch <= args.getLong("numEpochs"); epoch++)
	{ 
	  // add candidate learners and points to the populations
	  
	  genPop(pop, args, epoch, trainData._categories, action);
	  genPoints(trainData.size() - 1, PP, trainData);
	  
#ifdef CCM_DEBUG
	  cout << DEBUG << " EPOCH " << epoch;
	  for(int i = 0; i < pop._agents.size(); i++)
	    cout << " " << *pop._agents[i];
	  cout << endl;
	  
	  cout << DEBUG << " PP";
	  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
	    cout << " " << (*pIter)->id() << ":" << trainData.category((*pIter)->id());
	  cout << endl;
#endif
	  
	  // run all learners vs all points
	  
	  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
	    {
	      (*pIter)->clearOutcome(); // make sure clear outcomes
	      auction(pop, trainData, **pIter);
	    }
	  
	  // at this point points in PP have outcomes set wrt current population
	  // collect outcomes and distinctions

	  for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
	    delete oIter->second;
	  outcomes.clear();
	  
	  for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
	    delete dIter->second;
	  distincts.clear();

	  labels.clear();
	  
	  collectOutcomes(pop, PP, outcomes, labels); // collect outcomes
	  collectDistincts(PP, pop, distincts); // collect distinctions
	  
#ifdef CCM_DEBUG
	  for(int i = 0; i < pop.size(); i++)
	    {
	      cout << VECTS << " OUTCOMES " <<  epoch << " " << *pop[i];
	      for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++)
		{
		  if(j % 4 == 0) cout << " ";
		  cout << " ";
		  cout << (*outcomes[pop[i]->id()])[j];
		}
	      cout << endl;
	    }

	  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
	    {
	      cout << VECTS << " DISTINCTS " << epoch << " " << (*pIter)->id() << " " << (*pIter)->category();
	      for(int i = 0; i < pop.size(); i++)
		{
		  cout << " " << pop[i]->id();
		  for(int j = 0; j < pop.size(); j++)
		    {
		      if(j % 4 == 0) cout << " ";
		      index = i * pop.size() + j;
		      cout << (*distincts[(*pIter)->id()])[index];
		    }
		}
	      cout << endl;
	    }
#endif
	  
	  normalizeOutcomes(outcomes, labels, trainData.getCategory(action)); // normalize outcomes to account for global effect

#ifdef CCM_DEBUG
	  for(int i = 0; i < pop.size(); i++)
	    {
	      cout << VECTS << " NORM OUTCOMES " <<  epoch << " " << *pop[i];
	      for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++)
		{
		  if(j % 4 == 0) cout << " ";
		  cout << " ";
		  cout << (*outcomes[pop[i]->id()])[j];
		}
	      cout << endl;
	    }
#endif

	  selectPop(pop, outcomes, lDom);

	  selectPoints(PP, distincts, pCount, pDom);
	  
	  if(epoch % args.getLong("outputMod") == 0 || epoch == args.getLong("numEpochs"))
	    {	      
	      for(pIter = PP.begin(); pIter != PP.end(); pIter++)
		{
		  (*pIter)->clearOutcome(); // make sure clear outcomes
		  auction(pop, trainData, **pIter);
		}
	      
	      for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
		delete oIter->second;
	      outcomes.clear();

	      labels.clear();
	      
	      collectOutcomes(pop, PP, outcomes, labels); // collect the outcomes
	      
	      for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
		delete dIter->second;
	      distincts.clear();
	      
	      collectDistincts(PP, pop, distincts); // collect distinctions

	      double sumOut = 0;
	      for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++)
		for(int o = 0; o < oIter->second->size(); o++)
		  sumOut += (*(oIter->second))[o];

	      sumOut = sumOut / (pop.size() * PP.size());

	      double sumDist = 0;
	      for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
		for(int d = 0; d < dIter->second->size(); d++)
		  sumDist += (*(dIter->second))[d];
	      sumDist = sumDist / (pop.size() * pop.size() * PP.size());
	      

	      cout << STATS << " E" << epoch << " " << epoch <<  " ACTION " << action << " " << trainData._categories[action];
	      cout << " LDOM " << lDom << " PDOM " << pDom;
	      cout << " SUMOUT " << sumOut << " SUMDIST " << sumDist;

	      cout << " POINT";
	      for(cIter = pCount.begin(); cIter != pCount.end(); cIter++)
		cout << " " << cIter->first;
	      cout << " COUNT";
	      for(cIter = pCount.begin(); cIter != pCount.end(); cIter++)
		cout << " " << cIter->second;
	      cout << endl;

	      if(epoch == args.getLong("numEpochs")) // last epoch
		{		  
		  
		  cout.precision(2);
		  for(int i = 0; i < pop.size(); i++)
		    {
		      cout << VECTS << " OUTCOMES " <<  epoch << " " << *pop[i];

		      cout << showpoint;
		      for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++)
			{
			  if(j % 4 == 0 && j != 0) cout << " ";
			  cout << " ";
			  cout << (*outcomes[pop[i]->id()])[j];
			}
		      cout << endl;
		      cout << noshowpoint;
		    }
		  cout.precision(3);
		  
		  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
		    {
		      cout << VECTS << " DISTINCTS " << epoch <<  " " << (*pIter)->category() << " " << (*pIter)->id();
		      for(int i = 0; i < pop.size(); i++)
			{
			  cout << " ";
			  for(int j = 0; j < pop.size(); j++)
			    {
			      if(j % 4 == 0 && j != 0) cout << " ";
			      index = i * pop.size() + j;
			      cout << (*distincts[(*pIter)->id()])[index];
			    }
			}
		      cout << endl;
		    }
		}
	    }
	}
      
      // collect outcome wrt selected individuals and points
      
      for(pIter = PP.begin(); pIter != PP.end(); pIter++)
	{
	  (*pIter)->clearOutcome(); // make sure clear outcomes
	  auction(pop, trainData, **pIter);
	}
      
      for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
	delete oIter->second;
      outcomes.clear();

      labels.clear();
      
      collectOutcomes(pop, PP, outcomes, labels); // collect the outcomes

      normalizeOutcomes(outcomes, labels, trainData.getCategory(action)); // normalize outcomes to account for global effect
      reduce(pop, outcomes);
      
      Agent *a;

      cout.precision(2);
      while(pop.size() > 0)
	{
	  a = pop.removeLast();
	  
	  cout << PROGR << " REDUCE";
	  cout << showpoint;
	  for(int i = 0; i < outcomes[a->id()]->size(); i++)
	    {
	      if(i % 4 == 0 && i != 0) cout << " ";
	      cout << " " << (*outcomes[a->id()])[i];
	    }
	  cout << noshowpoint;
	  
	  if(a->active() == false)
	    {
	      cout << " DROP " << *a << endl;
	      delete a;
	    }
	  else
	    {
	      cout << " SAVE " << *a << endl;
	      solution.pushBack(a);
	    }
	}
      cout.precision(3);

      for(pIter == PP.begin(); pIter != PP.end(); pIter++)
	delete *pIter;

      for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
	delete oIter->second;
      
      for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
	delete dIter->second;
    }

    double sys2, usr2;
    if(!getusage(sys2, usr2))
      die(__FILE__, __FUNCTION__, __LINE__, "cannot get rusage");
    cout << "RUSAGE SYS1 = " << sys1 << " USR1 = " << usr1 << " SUM1 = " << (sys1 + usr1);
    cout << " SYS2 = " << sys2 << " USR2 " << usr2 << " SUM2 = " << (sys2 + usr2);
    cout << " TRAINTIME = " << (sys2 + usr2) - (sys1 + usr1) << endl;
    

    DataSet testData(args.getString("testDataFile"), args.getLong("numFeatures"), args.getLong("testNumPatterns"));
    //    Stats testStats(&testData);
    
    
    test(solution, trainData, testData, args);
    
    while(solution.size() > 0)
      delete solution.removeLast();
}

void test(Pop &pop, DataSet &trainData, DataSet &testData, Args &args)
{
  Stats trainStats(&trainData);
  Stats testStats(&testData);

  trainStats.update(pop, args.getLong("numEpochs"));
  testStats.update(pop, args.getLong("numEpochs"));
	  
  trainStats.print("main", true);
  testStats.print("main", false);

}

// Initialize population so that the bid vectors over PP are different
// for each agent. If the accuracy is used as an outcome on each dimension,
// this should result in an initial population of agents where the
// outcome vectors are not equal (diversity).
void initPop(int numAgents, 
	     Args &args,
	     Pop &pop, 
	     vector < int > &actions,
	     int action,
	     set < Point * > &PP,
	     DataSet &data)
{
  map < long, vector < double > * > bidMap;
  map < long, vector < double > * > :: iterator bIter;
  vector < double > *vd;
  
  double *REG = (double *) alloca(REGISTERS * sizeof(double));
  
  set < Point * > :: iterator pIter;
  
  long tries = 0;

  // Create the first learner and keep it as is.
  
  Agent *a;
  
  a = new Agent(args, 0, actions[action]);
  a->old(true);
  pop.pushBack(a);
  
  // Calculate the bid vector.
  
  vd = new vector < double >;
  
  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
    vd->push_back(a->bid(data.feature((*pIter)->id()), REG));
  
  bidMap.insert(map < long, vector < double > * > :: value_type(a->id(), vd));
  
  // Add more learners.

  while(pop.size() < numAgents)
    {
      a = new Agent(args, 0, actions[action]);
      
      vd = new vector < double >;
      
      for(pIter = PP.begin(); pIter != PP.end(); pIter++)
	vd->push_back(a->bid(data.feature((*pIter)->id()), REG));
      
      // Compare against existing learners' bid vectors.
      
      for(bIter = bidMap.begin(); bIter != bidMap.end(); bIter++)
	if(isEqual(*vd, *(bIter->second), 0.1))	break;
      
      if(bIter == bidMap.end()) // Add agent to population.
	{
	  a->old(true);
	  pop.pushBack(a);
	  bidMap.insert(map < long, vector < double > * > :: value_type(a->id(), vd));
	}
      else // Delete agent.
	{
	  delete vd;
	  delete a;
	}

      tries++;
    }

#ifdef CCM_DEBUG
  
  cout << INITI << " TRIES " << tries << endl;
  for(bIter = bidMap.begin(); bIter != bidMap.end(); bIter++)
    {
      cout << INITI << " " << bIter->first;
      for(int i = 0; i < bIter->second->size(); i++)
	cout << " " << (*(bIter->second))[i];
      cout << endl;
      delete bIter->second;
    }
#endif

}


void initPop(int numAgents, 
	     Args &args,
	     Pop &pop, 
	     vector < int > &actions,
	     int action)
{
  Agent *a;
  
  for(int i = 0; i < numAgents; i++)
    {
      a = new Agent(args, 0, actions[action]);
      a->old(true);
      pop.pushBack(a);
    }
  
#ifdef CCM_DEBUG
  cout << DEBUG << " INITPOP";
  for(int i = 0; i < pop.size(); i++)
    cout << " " << *pop[i];
  cout << endl;
#endif
}

void initPoints(int numPoints, 
		long maxId, // ids must be from 0 to maxId inclusive
		set < Point * > &P, // where to store the new points
		map < int, int > &pCount, // category count matrix with keys present
		DataSet &data)
{
  if(numPoints > maxId + 1 || pCount.size() > maxId + 1)
    die(__FILE__, __FUNCTION__, __LINE__, "cannot generate that many points initially");

  if(numPoints < pCount.size())
    die(__FILE__, __FUNCTION__, __LINE__, "need more points initially to get one of each class");
  
  // generate new set of ids
  
  set < long > ids;
  long newPoint;

//   int numFixed = (int) numPoints / pCount.size();
//   int numRand = numPoints % pCount.size();

//  // still possible to enter infinite loop below if not enough points of EACH class
  
//   map < int, int > :: iterator pIter;
//   for(int i = 0; i < numFixed; i++)
//     for(pIter = pCount.begin(); pIter != pCount.end(); pIter++)
//       {
// 	newPoint = (long) (drand48() * (maxId + 1));
// 	while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
// 	  newPoint = (newPoint + 1) % (maxId + 1);
//       }

//   for(int i = 0; i < numRand; i++)
//     {
//       newPoint = (long) (drand48() * (maxId + 1));
//       while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
// 	newPoint = (newPoint + 1) % (maxId + 1);
//     }
  
  // make sure get at least one of each point category
  
  map < int, int > :: iterator pIter;
  for(pIter = pCount.begin(); pIter != pCount.end(); pIter++)
    {
      newPoint = (long) (drand48() * (maxId + 1));
      while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
	newPoint = (newPoint + 1) % (maxId + 1);
    }
  
  while(ids.size() < numPoints)
    {
      newPoint = (long) (drand48() * (maxId + 1));
      
      while((ids.insert(newPoint)).second == false)
	newPoint = (newPoint + 1) % (maxId + 1);
    }
  
  // create new points using ids
  
  set < long > :: iterator newIter;
  Point *p;
  
#ifdef CCM_DEBUG
  cout << DEBUG << " INITPOINTS";
#endif
  for(newIter = ids.begin(); newIter != ids.end(); newIter++)
    {
      p = new Point(*newIter, data.category(*newIter));
      p->old(true);
      P.insert(p);
      pCount[p->category()]++;
      
#ifdef CCM_DEBUG
      cout << " " << p->id();
#endif
    }
  
#ifdef CCM_DEBUG
  cout << endl;
#endif
}

void genPop(Pop &pop, Args &args, long epoch, vector < int > &actions, int action)
{
  int popSize = pop.size();
  
  Agent *a;
  
  double pam = -1; // probability of action mutation
  
  for(int i = 0; i < popSize; i++)
    {
      if(drand48() < 0.9) // generate from parent
	a = new Agent(args, epoch, *pop[i], actions, pam);
      else // generate new random
	a = new Agent(args, epoch, actions[action]);
      a->old(false);
      pop.pushBack(a);
    }
}

void genPoints(long maxId, // ids must be from 0 to maxId inclusive
	       set < Point * > &PP,
	       DataSet &data)
{
  int ppSize = PP.size();
  
  if(2 * ppSize > maxId + 1)
    die(__FILE__, __FUNCTION__, __LINE__, "cannot generate that many points");

  // get the ids of the existing points
  
  set < Point * > :: iterator oldIter;
  set < long  > oldIds;

  for(oldIter = PP.begin(); oldIter != PP.end(); oldIter++)
    oldIds.insert((*oldIter)->id());
  
  // generate new set of ids

  set < long > ids;
  long newPoint;
  
  while(ids.size() < ppSize)
    {
      newPoint = (long) (drand48() * (maxId + 1));
      
      while((oldIds.insert(newPoint)).second == false)
	newPoint = (newPoint + 1) % (maxId + 1);
      
      ids.insert(newPoint); // only insert if not in set PP
    }

  // create new points using ids

  set < long > :: iterator newIter;
  
  Point *p;
  
  for(newIter = ids.begin(); newIter != ids.end(); newIter++)
    {
      p = new Point(*newIter, data.category(*newIter));
      p->old(false);
      PP.insert(p);
    }
}

// gather outcomes so that it is more efficient when accessing them using learner id
void collectOutcomes(Pop &pop, set < Point * > &P, map < long , vector < double > * > &outcomes, vector < int > &labels)
{
  map < long, vector < double > * > :: iterator mIter;

  // add new outcomes assuming outcomes is clear

  if(outcomes.empty() == false)
    die(__FILE__, __FUNCTION__, __LINE__, "outcomes not cleared");

  if(labels.empty() == false)
    die(__FILE__, __FUNCTION__, __LINE__, "labels not cleared");
  
  for(int i = 0; i < pop.size(); i++)
    outcomes.insert(map < long, vector < double > * > :: value_type(pop[i]->id(), new vector < double >));
  
  set < Point * > :: iterator sIter;

  for(sIter = P.begin(); sIter != P.end(); sIter++) // for each point
    {
      labels.push_back((*sIter)->category());
      
      for(mIter = outcomes.begin(); mIter != outcomes.end(); mIter++) // for each learner
	(mIter->second)->push_back((*sIter)->getOutcome(mIter->first)); // push back outcome on the point
    }
}

void collectDistincts(set < Point * > &PP, Pop &pop, map < long, vector < int > * > &distincts)
{
  map < long, vector < int * > > :: iterator dIter;
  
  if(distincts.empty() == false)
    die(__FILE__, __FUNCTION__, __LINE__, "distinctions not cleared");
  
  set < Point * > :: iterator sIter;
  
  int popSize = pop.size();
  double out1, out2;
  map < long, vector < int > * > :: iterator mIter;
  
  for(sIter = PP.begin(); sIter != PP.end(); sIter++)
    {
      mIter = (distincts.insert(map < long, vector < int > * > :: value_type((*sIter)->id(), new vector < int >))).first;
      
      for(int i = 0; i < popSize; i++)
	{
	  out1 = (*sIter)->getOutcome(pop[i]->id());

	  for(int j = 0; j < popSize; j++) // should consider only distinctions between member of the same class?
	    {
	      out2 = (*sIter)->getOutcome(pop[j]->id());
	      
	      //	      if(isEqual(out1, out2))
	      if(isEqual(out1, out2, 0.1))
		mIter->second->push_back(0);
	      else if(out1 > out2)
		mIter->second->push_back(1);
	      else // out1 < out 2
		mIter->second->push_back(0);
	    }
	}
    }
}

// Assumes outcomes in [0, 1] and denote accuracy with 1 the most accurate.
// assume all learners of action 'action'
void normalizeOutcomes(map < long, vector < double > * > &outcomes, vector < int > &labels, int action)
{  
  map < long, vector < double > * > :: iterator iter;
  vector < double > *vd;
  double sumOutIn, sumOutOut;
  int sumCntIn, sumCntOut;
  double muIn, muOut;

  sumCntIn = sumCntOut = 0; // get counts of in and out of class exemplars
  for(int i = 0; i < labels.size(); i++)
    if(labels[i] == action) sumCntIn++;
    else sumCntOut++;
  
  for(iter = outcomes.begin(); iter != outcomes.end(); iter++)
    {
      vd = iter->second; // get outcome vector for this learner
      
      sumOutIn = sumOutOut = 0; // reset sums out outcomes
      
      if(vd->size() != labels.size())
 	die(__FILE__, __FUNCTION__, __LINE__, "outcome vector does not match label vector");
      
      for(int i = 0; i < vd->size(); i++)
	{
	  if(labels[i] == action) sumOutIn += (*vd)[i];
	  else sumOutOut += (*vd)[i];
	}

      muIn = (1 + sumOutIn) / (1 + sumCntIn);
      muOut = (1 + sumOutOut) / (1 + sumCntOut);
      
      for(int i = 0; i < vd->size(); i++)
	{
	  if(labels[i] == action) (*vd)[i] = (*vd)[i] * muOut;
	  else (*vd)[i] = (*vd)[i] * muIn;
	}
    }
  

//   map < long, vector < double > * > :: iterator iter;
//   vector < double > *vd;
//   double sumOut;
//   int sumCnt;
//   double mu;
  
//   for(iter = outcomes.begin(); iter != outcomes.end(); iter++) // consider each outcome vector
//     {
//       vd = iter->second; // get the outcome vector
//       sumOut = 0;
//       sumCnt = 0;
      
//       if(vd->size() != labels.size())
// 	die(__FILE__, __FUNCTION__, __LINE__, "outcome vector does not match label vector");

//       for(int i = 0; i < vd->size(); i++) // consider each outcome
// 	{
// 	  if(labels[i] == action) continue; // skip outcome if for matching pattern
	  
// 	  sumOut += (*vd)[i];
// 	  sumCnt += 1;
// 	}

//       mu = (1 + sumOut) / (1 + sumCnt);

//       for(int i = 0; i < vd->size(); i++)
// 	{
// 	  (*vd)[i] = (*vd)[i] * mu; // normalize by mean accuracy
// 	}
//     }

//   map < long, vector < double > * > :: iterator iter;
//   vector < double > *vd;
//   double sumOut;

//   for(iter = outcomes.begin(); iter != outcomes.end(); iter++)
//     {
//       vd = iter->second;
//       sumOut = 0;

//       if(vd.size() != labels.size())
// 	die(__FILE__, __FUNCTION__, __LINE__, "outcome vector does not match label vector");
      
//       for(int i = 0; i < vd->size(); i++)
// 	{
// 	  sumOut += (*vd)[i];
// 	}
      
//       sumOut = sumOut / (1 + vd->size()); // mean accuracy

//       for(int i = 0; i < vd->size(); i++) (*vd)[i] = (*vd)[i] * sumOut; // normalize by mean accuracy
//     }

}

bool dominated(vector < double > &is, vector < double > &by)
{
  bool equal = true;
  
  if(is.size() != by.size())
    die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
  for(int i = 0; i < is.size(); i++)
    {
      if(isEqual(is[i], by[i]) == true) continue;
      
      equal = false;
      
      if(is[i] > by[i]) return false;
    }
  
  return equal == false;
}

bool dominated(vector < double > &is, vector < double > &by, double e)
{
  bool equal = true;
  
  if(is.size() != by.size())
    die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
  for(int i = 0; i < is.size(); i++)
    {
      if(isEqual(is[i], by[i], e) == true) continue;
      
      equal = false;
      
      if(is[i] > by[i]) return false;
    }
  
  return equal == false;
}

// 'is' and 'by' are equal -> return 0
// 'is' dominated by 'by' -> return 1
// otherwise -> return -1
int dominatedOrEqual(vector < double > &is, vector < double > &by, double e)
{
  bool equal = true;
  
  if(is.size() != by.size())
    die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
  for(int i = 0; i < is.size(); i++)
    {
      if(isEqual(is[i], by[i], e) == true) continue;
      
      equal = false;
      
      if(is[i] > by[i]) return -1; // 'is' not dominated by 'by'
    }
  
  if(equal == true) return 0;

  // not equal and 'is' <= 'by' on all objectives
  // therefore 'by' > 'is' on at least one objective and 'by' never worse than 'is'

  return 1; // 'is' dominated by 'by'
}

bool dominated(vector < int > &is, vector < int > &by)
{
  bool equal = true;
  
  if(is.size() != by.size())
    die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
  for(int i = 0; i < is.size(); i++)
    {
      if(is[i] == by[i]) continue;
      
      equal = false;

      if(is[i] > by[i]) return false;
    }
  
  return equal == false;
}

Agent * auction(Pop &pop, DataSet &data, Point &point)
{
  if(pop._agents.size() < 2)
    die(__FILE__, __FUNCTION__, __LINE__, "not enough agents to auction");

  double *instance = data.feature(point.id());
  int outcome = data.category(point.id());
  
  double due; // what winner must pay out
  double reward; // what winner receives
  
  double *REG = (double *) alloca(REGISTERS * sizeof(double));
  
  Agent *winner = *(pop.begin());
  winner->_dummy = winner->bid(instance, REG);

  vector < Agent * > :: iterator iter = pop.begin();
  
  for(iter++; iter != pop.end(); iter++)
    {
      (*iter)->_dummy = (*iter)->bid(instance, REG);
      
      if((*iter)->outbids(*winner))
	winner = *iter;
    }
  
  // at this point _dummy is the bid of each agent and winner points to the highest bidder
  // the outcomes should be cleared for the point
  
  setOutcome(point, pop, outcome, winner);
  
  due = winner->_dummy;
  
  if(outcome == winner->action())
    reward = 1;
  else
    reward = 0;
  
  winner->_W += (reward - due);
  winner->recordWin(due, reward);
  
#ifdef CCM_DEBUG
  cout << DEBUG << " PAT " << point.id() << " CAT " << outcome << " WINNER " << *winner << " BID " << winner->_dummy << " PAY " << due << endl;
#endif

  return winner;
}

void setOutcome(Point &point, Pop &pop, int outcome, Agent *winner)
{  
  for(int i = 0; i < pop.size(); i++) // THIS CAN BE SPED UP IF REWARD IS EITHER 1 0R 0
    {
      if(pop[i]->action() == outcome)
	point.setOutcome(pop[i]->id(), pop[i]->_dummy); // 1 - (1 - pop[i]->_dummy)
      else
	point.setOutcome(pop[i]->id(), 1.0 - pop[i]->_dummy); // 1 - (0 + pop[i]->_dummy)
    }

			
//   Agent *aPtr = NULL; // runner up
  
// #ifdef CCM_DEBUG
//   cout << DEBUG;
// #endif
  
//   for(int i = 0; i < pop.size(); i++) // consider each agent
//     {
// #ifdef CCM_DEBUG
//       cout << " " << *pop[i] << ">>" << pop[i]->_dummy;
// #endif 

//       point.setOutcome(pop[i]->id(), 1);
      
//       if(pop[i]->action() != winner->action()) // class differet from winner
// 	{
// 	  if((aPtr == NULL) || (pop[i]->outbids(*aPtr) == true))
// 	    aPtr = pop[i];
// 	}
//     }

// #ifdef CCM_DEBUG
//   cout << endl;
//   if(aPtr != NULL)
//     cout << DEBUG << " >>APTR" << *aPtr << " BID " << aPtr->_dummy << endl;
//   else
//     cout << DEBUG << " >>APTRNULL" << endl;
// #endif
    
//   for(int i = 0; i < pop.size(); i++)
//     {
//       if((pop[i]->action() == winner->action()) && ((aPtr == NULL) || (pop[i]->outbids(*aPtr))))
// 	{
// 	  if(pop[i]->action() == outcome) 
// 	    point.setOutcome(pop[i]->id(), 2);
// 	  else
// 	    point.setOutcome(pop[i]->id(), 0);
// 	}
//     }
}

void selectPop(Pop &pop, map < long, vector < double > * > &outcomes, long &lDom)
{
  vector < Agent * > parent;
  vector < Agent * > child;
  
  for(int i = 0; i < pop.size(); i++)
    {
      if(pop[i]->old() == true)
	parent.push_back(pop[i]);
      else
	child.push_back(pop[i]);
    }
  
  pop.clear();
  
  random_shuffle(child.begin(), child.end());
  
  int pSize = parent.size();
  int cSize = child.size();

  bool dom; // april19
  //  int dom; // april19
  int p, c;

  // remove children with duplicate outcome vectors

  for(int i = 0; i < child.size(); i++)
    {
      child[i]->active(true);
      
      for(int j = 0; j < i; j++)
	{
	  if(child[j]->active() == false) continue;

	  if(isEqual(*outcomes[child[i]->id()], *outcomes[child[j]->id()], 0.1) == true)
	    {
	      child[i]->active(false);
	      break;
	    }
	}
    }
  
  for(c = 0; c < cSize; c++)
    {
      if(child[c]->active() == false) continue; // There is a duplicate child.

      // Check to make sure child is not the same as a parent.
      for(p = 0; p < pSize; p++)
	{
	  if(isEqual(*outcomes[parent[p]->id()], *outcomes[child[c]->id()], 0.1) == true)
	    break;
	}
      if(p < pSize) continue;
      
      random_shuffle(parent.begin(), parent.end());
      
      for(p = 0; p < pSize; p++)
	{
	  if(parent[p]->old() == false) continue;

	  //	  dom = dominated(*outcomes[parent[p]->id()], *outcomes[child[c]->id()]);
	  dom = dominated(*outcomes[parent[p]->id()], *outcomes[child[c]->id()], 0.1); // april19
	  //dom = dominatedOrEqual(*outcomes[parent[p]->id()], *outcomes[child[c]->id()], 0.1); // april19
	  
	  if(dom == true) // april19
	  //if((dom == 1) || (dom == 0 && (child[c]->size() <= parent[p]->size()))) // april19
	    {
	      parent[p]->old(false);
	      child[c]->old(true);
	      
#ifdef CCM_DEBUG
	      cout << DEBUG << " " << *child[c] << " LREPLACES " << *parent[p] << endl;
#endif
	      lDom++;

	      break;
	    }
	}
    }
  
  for(int i = 0; i < pSize; i++)
    if(parent[i]->old() == true) pop.pushBack(parent[i]);
    else delete parent[i];
  
  for(int i = 0; i < cSize; i++)
    if(child[i]->old() == true) pop.pushBack(child[i]);
    else delete child[i];
}

void selectPoints(set < Point * > &PP, map < long, vector < int > * > &distincts, map < int, int > &pCount, long &pDom)
{
  vector < Point * > parent;
  vector < Point * > child;
  
  set < Point * > ::iterator pIter;
  
  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
    {
      if((*pIter)->old() == true)
	parent.push_back(*pIter);
      else
	child.push_back(*pIter);
    }
  
  PP.clear();
  
  random_shuffle(child.begin(), child.end());
  
  int pSize = parent.size();
  int cSize = child.size();
  
  bool dom;
  int c, p;

  // remove children with duplicate outcome vectors

  for(int i = 0; i < child.size(); i++)
    {
      child[i]->active(true);
      
      for(int j = 0; j < i; j++)
	{
	  if(child[j]->active() == false) continue;
	  
	  if(isEqual(*distincts[child[i]->id()], *distincts[child[j]->id()]) == true)
	    {
	      child[i]->active(false);
	      break;
	    }
	}
    }
  
  for(c = 0; c < cSize; c++)
    {
      if(child[c]->active() == false) continue; // There is a duplicate child.

      // Check to make sure child is not the same as a parent.
      for(p = 0; p < pSize; p++)
	{
	  if(isEqual(*distincts[parent[p]->id()], *distincts[child[c]->id()]) == true)
	    break;
	}
      if(p < pSize) continue;
      
      random_shuffle(parent.begin(), parent.end());
      
      for(p = 0; p < pSize; p++)
	{
	  if(parent[p]->old() == false || (pCount[parent[p]->category()] < 2 && parent[p]->category() != child[c]->category())) continue;
	  //	  if(parent[p]->old() == false || parent[p]->category() != child[c]->category()) continue;
	  
	  dom = dominated(*distincts[parent[p]->id()], *distincts[child[c]->id()]);
	  
	  if(dom == true)
	    {
	      parent[p]->old(false);
	      child[c]->old(true);
	      
	      pCount[parent[p]->category()]--;
	      pCount[child[c]->category()]++;
	      
#ifdef CCM_DEBUG
	      cout << DEBUG << " " << child[c]->id() << " PREPLACES " << parent[p]->id() << endl;
#endif
	      pDom++;

	      break;
	    }
	}
    }

  for(int i = 0; i < pSize; i++)
    if(parent[i]->old() == true) PP.insert(parent[i]);
    else delete parent[i];
  
  for(int i = 0; i < cSize; i++)
    if(child[i]->old() == true) PP.insert(child[i]);
    else delete child[i];

}

void reduce(Pop &pop, map < long, vector < double > * > &outcomes)
{
  for(int i = 0; i < pop.size(); i++) pop[i]->active(true);
  
  for(int i = 0; i < pop.size(); i++)
    {
      //      pop[i]->active(true);
      
      // Set to false if outcome vectors the same.

      for(int j = 0; j < i; j++)
	{
	  if((pop[j]->active() == true) && (isEqual(*outcomes[pop[i]->id()], *outcomes[pop[j]->id()], 0.1) == true))
	    {
	      pop[i]->active(false);
	      break;
	    }
	}
      if(pop[i]->active() == false) continue;

      // Set to false if dominated.
      
      for(int j = 0; j < pop.size(); j++)
	{
	  if(pop[i]->action() != pop[j]->action()) continue;
	  if(pop[j]->active() == false) continue;
	  
	  if(i != j)
	    {
	      //	      if(dominated(*outcomes[pop[i]->id()], *outcomes[pop[j]->id()]) == true)
	      if(dominated(*outcomes[pop[i]->id()], *outcomes[pop[j]->id()], 0.1) == true)
		{
		  pop[i]->active(false);
		  break;
		}
	    }
	}
    }
}







// Proposal 43


// typedef struct
// {
//   long learnerCount;
//   long pointCount;
// } stats;

// // with SWB the learner can always bid max and not be penalized for getting
// // something wrong and always be in 1st front

// void train(Args &);

// void initPop(int, Args &, Pop &, vector < int > &, map < int, int > &); // initialize random learner population
// void initPoints(int, long, set < Point * > &, map < int, int > &, DataSet &); // initialize random point population

// void genPop(Pop &, Args &, long, vector < int > &);
// void genPoints(long, set < Point * > &, DataSet &);

// void auction(Pop &, DataSet &, Point &);
// void setOutcome(Point &, Pop &, int, Agent *);

// void collectOutcomes(Pop &, set < Point * > &, map < long , vector < double > * > &);
// void collectDistincts(set < Point * > &, Pop &, map < long, vector < int > * > &);

// bool dominated(vector < double > &, vector < double > &);
// bool dominated(vector < double > &, vector < double > &, double);
// bool dominated(vector < int > &, vector < int > &);

// void selectPop(Pop &, map < long, vector < double > * > &, map < int, int > &, stats &);
// void selectPoints(set < Point * > &, map < long, vector < int > * > &, map < int, int > &, stats &);

// void reduce(Pop &, map < long, vector < double > * > &);

// // int fnds(map < long, vector < double > * > &, map < long, int > &); // fast non-dominated sort

// int main(int argc, char **argv)
// {
//   if(argc != 2)
//     die(__FILE__, __FUNCTION__, __LINE__, "parameter file not specified");
  
//   Args args(argv[1]);
//   cout << args << endl;
//   train(args);
  
//   struct rusage res;
//   if(getrusage(RUSAGE_SELF, &res) == -1)
//     {
//       cout << "RUSAGE ERROR";
//     }
//   else
//     {
//       double sys = (double) res.ru_stime.tv_sec + res.ru_stime.tv_usec * 1e-6;
//       double usr = (double) res.ru_utime.tv_sec + res.ru_utime.tv_usec * 1e-6;
      
//       cout << "RUSAGE SYS " << sys << " USR " << usr << " SUM " << sys + usr << endl;
//     }
  
//   return 0; 
// }

// void train(Args &args)
// {
//   srand48(args.getLong("seed"));
  
//   DataSet trainData(args.getString("trainDataFile"), args.getLong("numFeatures"), args.getLong("trainNumPatterns"));
//   DataSet testData(args.getString("testDataFile"), args.getLong("numFeatures"), args.getLong("testNumPatterns"));
  
//   Stats trainStats(&trainData);
//   Stats testStats(&testData);
  
//   long epoch;
  
//   Pop pop;
//   vector < Agent * >::iterator iter;

//   set < Point * > PP; // point population
//   set < Point * >::iterator pIter; // point populatin iterator
  
//   map < long, vector < double > * > outcomes; // (learner id) -> (outcome vector)
//   map < long, vector < double > * > :: iterator oIter;
  
//   map < long, vector < int > * > distincts; // (point id) -> (distinction vector)
//   map < long, vector < int > * > :: iterator dIter;
  
//   map < int, int > pCount; // count patterns of each class
//   map < int, int > aCount; // count agents of each action
//   map < int, int > :: iterator cIter;

//   stats s;
//   s.learnerCount = 0;
//   s.pointCount = 0;

//   int index;
  
//   for(int i = 0; i < trainData._categories.size(); i++)
//     {
//       pCount.insert(map < int, int > :: value_type(trainData._categories[i], 0));
//       aCount.insert(map < int, int > :: value_type(trainData._categories[i], 0));
//     }
  
//   // initialize populations
//   initPop(args.getLong("numAgents"), args, pop, trainData._categories, aCount);
//   initPoints(args.getLong("numTests"), trainData.size() - 1, PP, pCount, trainData);
  
//   // start epoch
//   for(epoch = 0; epoch <= args.getLong("numEpochs"); epoch++)
//     { 
//       // add candidate learners and points to the populations
      
//       genPop(pop, args, epoch, trainData._categories);
//       genPoints(trainData.size() - 1, PP, trainData);

// #ifdef CCM_DEBUG
//       cout << DEBUG << " EPOCH " << epoch;
//       for(int i = 0; i < pop._agents.size(); i++)
// 	cout << " " << *pop._agents[i];
//       cout << endl;
      
//       cout << DEBUG << " PP";
//       for(pIter = PP.begin(); pIter != PP.end(); pIter++)
// 	cout << " " << (*pIter)->id() << ":" << trainData.category((*pIter)->id());
//       cout << endl;
// #endif
      
//       // run all learners vs all points

//       for(pIter = PP.begin(); pIter != PP.end(); pIter++)
// 	{
// 	  (*pIter)->clearOutcome(); // make sure clear outcomes
// 	  auction(pop, trainData, **pIter);
// 	}
      
//       // at this point points in PP have outcomes set wrt current population
//       // collect outcomes and distinctions
      
//       collectOutcomes(pop, PP, outcomes); // collect the outcomes first
//       collectDistincts(PP, pop, distincts); // collect distinctions
      
// #ifdef CCM_DEBUG
//       for(int i = 0; i < pop.size(); i++)
// 	{
// 	  cout << DEBUG << " OUTCOMES " <<  epoch << " " << *pop[i];
// 	  for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++)
// 	    {
// 	      if(j % 4 == 0) cout << " ";
// 	      cout << " ";
// 	      cout << (*outcomes[pop[i]->id()])[j];
// 	    }
// 	  cout << endl;
// 	}

//       for(pIter = PP.begin(); pIter != PP.end(); pIter++)
// 	{
// 	  cout << DEBUG << " DISTINCTS " << epoch << " " << (*pIter)->id();
// 	  for(int i = 0; i < pop.size(); i++)
// 	    {
// 	      cout << " " << pop[i]->id();
// 	      for(int j = 0; j < pop.size(); j++)
// 		{
// 		  if(j % 4 == 0) cout << " ";
// 		  index = i * pop.size() + j;
// 		  cout << (*distincts[(*pIter)->id()])[index];
// 		}
// 	    }
// 	  cout << endl;
// 	}
// #endif
      
//       selectPop(pop, outcomes, aCount, s);
//       selectPoints(PP, distincts, pCount, s);
      
//       if(epoch % args.getLong("outputMod") == 0 || epoch == args.getLong("numEpochs"))
// 	{
// 	  // collect outcome wrt selected individuals and points
	  
// 	  for(pIter = PP.begin(); pIter != PP.end(); pIter++)
// 	    {
// 	      (*pIter)->clearOutcome(); // make sure clear outcomes
// 	      auction(pop, trainData, **pIter);
// 	    }
	  
// 	  for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
// 	    delete oIter->second;
// 	  outcomes.clear();
	  
// 	  collectOutcomes(pop, PP, outcomes); // collect the outcomes
	  
// 	  reduce(pop, outcomes);
	  
// 	  trainStats.update(pop, epoch);
// 	  testStats.update(pop, epoch);
	  
// 	  trainStats.print("main", true);
// 	  testStats.print("main", false);

// 	  cout << "COUNT " << epoch;
// 	  cout << " LEARN";
// 	  for(cIter = aCount.begin(); cIter != aCount.end(); cIter++)
// 	    cout << " " << cIter->first;
// 	  cout << " COUNT";
// 	  for(cIter = aCount.begin(); cIter != aCount.end(); cIter++)
// 	    cout << " " << cIter->second;
// 	  cout << " POINT";
// 	  for(cIter = pCount.begin(); cIter != pCount.end(); cIter++)
// 	    cout << " " << cIter->first;
// 	  cout << " COUNT";
// 	  for(cIter = pCount.begin(); cIter != pCount.end(); cIter++)
// 	    cout << " " << cIter->second;
// 	  cout << " learnerCount " << s.learnerCount << " pointCount " << s.pointCount << endl;
	  
// 	  if(epoch == args.getLong("numEpochs")) // last epoch
// 	    {
// 	      for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
// 		delete dIter->second;
// 	      distincts.clear();

// 	      collectDistincts(PP, pop, distincts); // collect distinctions
	      
// 	      for(int i = 0; i < pop.size(); i++)
// 		{
// 		  cout << DEBUG << " OUTCOMES " <<  epoch << " " << *pop[i];
// 		  for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++)
// 		    {
// 		      if(j % 4 == 0) cout << " ";
// 		      cout << " ";
// 		      cout << (*outcomes[pop[i]->id()])[j];
// 		    }
// 		  cout << endl;
// 		}
	      
// 	      for(pIter = PP.begin(); pIter != PP.end(); pIter++)
// 		{
// 		  cout << DEBUG << " DISTINCTS " << epoch << " " << (*pIter)->id();
// 		  for(int i = 0; i < pop.size(); i++)
// 		    {
// 		      //    cout << " " << pop[i]->id();
// 		      cout << " ";
// 		      for(int j = 0; j < pop.size(); j++)
// 			{
// 			  if(j % 4 == 0) cout << " ";
// 			  index = i * pop.size() + j;
// 			  cout << (*distincts[(*pIter)->id()])[index];
// 			}
// 		    }
// 		  cout << endl;
// 		}
// 	    }
// 	}
      
//       for(oIter = outcomes.begin(); oIter != outcomes.end(); oIter++) 
// 	delete oIter->second;
//       outcomes.clear();

//       for(dIter = distincts.begin(); dIter != distincts.end(); dIter++)
// 	delete dIter->second;
//       distincts.clear();
//     }

//   // free memory
  
//   while(pop._agents.size() > 0)
//     {
//       delete pop._agents.back();
//       pop._agents.pop_back();
//     }
// }

// void initPop(int numAgents, 
// 	     Args &args,
// 	     Pop &pop, 
// 	     vector < int > &actions,
// 	     map < int, int > &aCount)
// {
//   Agent *a;
  
//   if(numAgents < actions.size())
//     die(__FILE__, __FUNCTION__, __LINE__, "not enough agents");
  
//   int numFixed = (int) numAgents / actions.size();
//   int numRand = numAgents % actions.size();
  
//   // assign equal number of individuals until no longer can
  
//   for(int i = 0; i < numFixed; i++)
//     for(int j = 0; j < actions.size(); j++)
//       {
// 	a = new Agent(args, 0, actions[j]);
// 	a->old(true);
// 	pop.pushBack(a);
// 	aCount[a->action()]++;
//       }
  
//   // randomly assign remainder of individuals
  
//   for(int i = 0; i < numRand; i++)
//     {
//       a = new Agent(args, 0, actions[(int) (drand48() * actions.size())]);
//       a->old(true);
//       pop.pushBack(a);
//       aCount[a->action()]++;
//     }
  
// #ifdef CCM_DEBUG
//   cout << DEBUG << " INITPOP";
//   for(int i = 0; i < pop.size(); i++)
//     cout << " " << *pop[i];
//   cout << endl;
// #endif
// }

// void initPoints(int numPoints, 
// 		long maxId, // ids must be from 0 to maxId inclusive
// 		set < Point * > &P, // where to store the new points
// 		map < int, int > &pCount, // category count matrix with keys present
// 		DataSet &data)
// {
//   if(numPoints > maxId + 1 || pCount.size() > maxId + 1)
//     die(__FILE__, __FUNCTION__, __LINE__, "cannot generate that many points initially");

//   if(numPoints < pCount.size())
//     die(__FILE__, __FUNCTION__, __LINE__, "need more points initially");
  
//   // generate new set of ids
  
//   set < long > ids;
//   long newPoint;

//   int numFixed = (int) numPoints / pCount.size();
//   int numRand = numPoints % pCount.size();

//   map < int, int > :: iterator pIter;
//   for(int i = 0; i < numFixed; i++)
//     for(pIter = pCount.begin(); pIter != pCount.end(); pIter++)
//       {
// 	newPoint = (long) (drand48() * (maxId + 1));
// 	while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
// 	  newPoint = (newPoint + 1) % (maxId + 1);
//       }

//   for(int i = 0; i < numRand; i++)
//     {
//       newPoint = (long) (drand48() * (maxId + 1));
//       while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
// 	newPoint = (newPoint + 1) % (maxId + 1);
//     }
  
// //   // make sure get at least one of each point category
  
// //   map < int, int > :: iterator pIter;
// //   for(pIter = pCount.begin(); pIter != pCount.end(); pIter++)
// //     {
// //       newPoint = (long) (drand48() * (maxId + 1));
// //       while((data.category(newPoint) != pIter->first) || ((ids.insert(newPoint)).second == false))
// // 	newPoint = (newPoint + 1) % (maxId + 1);
// //     }
  
// //   while(ids.size() < numPoints)
// //     {
// //       newPoint = (long) (drand48() * (maxId + 1));
      
// //       while((ids.insert(newPoint)).second == false)
// // 	newPoint = (newPoint + 1) % (maxId + 1);
// //     }
  
//   // create new points using ids
  
//   set < long >::iterator newIter;
//   Point *p;
  
// #ifdef CCM_DEBUG
//   cout << DEBUG << " INITPOINTS";
// #endif
//   for(newIter = ids.begin(); newIter != ids.end(); newIter++)
//     {
//       p = new Point(*newIter, data.category(*newIter));
//       p->old(true);
//       P.insert(p);
//       pCount[p->category()]++;
      
// #ifdef CCM_DEBUG
//       cout << " " << p->id();
// #endif
//     }
  
// #ifdef CCM_DEBUG
//   cout << endl;
// #endif
// }

// void genPop(Pop &pop, Args &args, long epoch, vector < int > &actions)
// {
//   int popSize = pop.size();
  
//   Agent *a;

//   double pam = args.getDouble("pActionMutate");
  
//   for(int i = 0; i < popSize; i++)
//     {
//       if(drand48() < 0.9) // generate from parent
// 	a = new Agent(args, epoch, *pop[i], actions, pam);
//       else // generate new random
// 	a = new Agent(args, epoch, actions[(int) (drand48() * actions.size())]);
//       a->old(false);
//       pop.pushBack(a);
//     }
// }

// void genPoints(long maxId, // ids must be from 0 to maxId inclusive
// 	       set < Point * > &PP,
// 	       DataSet &data)
// {
//   int ppSize = PP.size();
  
//   if(2 * ppSize > maxId + 1)
//     die(__FILE__, __FUNCTION__, __LINE__, "cannot generate that many points");

//   // get the ids of the existing points
  
//   set < Point * > :: iterator oldIter;
//   set < long  > oldIds;

//   for(oldIter = PP.begin(); oldIter != PP.end(); oldIter++)
//     oldIds.insert((*oldIter)->id());
  
//   // generate new set of ids

//   set < long > ids;
//   long newPoint;
  
//   while(ids.size() < ppSize)
//     {
//       newPoint = (long) (drand48() * (maxId + 1));
      
//       while((oldIds.insert(newPoint)).second == false)
// 	newPoint = (newPoint + 1) % (maxId + 1);
      
//       ids.insert(newPoint); // only insert if not in set PP
//     }

//   // create new points using ids

//   set < long > :: iterator newIter;
  
//   Point *p;
  
//   for(newIter = ids.begin(); newIter != ids.end(); newIter++)
//     {
//       p = new Point(*newIter, data.category(*newIter));
//       p->old(false);
//       PP.insert(p);
//     }
// }

// // gather outcomes so that it is more efficient when accessing them using learner id
// void collectOutcomes(Pop &pop, set < Point * > &P, map < long , vector < double > * > &outcomes)
// {
//   map < long, vector < double > * > :: iterator mIter;

//   // add new outcomes assuming outcomes is clear

//   if(outcomes.empty() == false)
//     die(__FILE__, __FUNCTION__, __LINE__, "outcomes not cleared");
  
//   for(int i = 0; i < pop.size(); i++)
//     outcomes.insert(map < long, vector < double > * > :: value_type(pop[i]->id(), new vector < double >));
  
//   set < Point * > :: iterator sIter;
  
//   for(sIter = P.begin(); sIter != P.end(); sIter++)
//     {
//       for(mIter = outcomes.begin(); mIter != outcomes.end(); mIter++)
// 	(mIter->second)->push_back((*sIter)->getOutcome(mIter->first));
//     }

// }

// void collectDistincts(set < Point * > &PP, Pop &pop, map < long, vector < int > * > &distincts)
// {
//   map < long, vector < int * > > :: iterator dIter;
  
//   if(distincts.empty() == false)
//     die(__FILE__, __FUNCTION__, __LINE__, "distinctions not cleared");
  
//   set < Point * > :: iterator sIter;
  
//   int popSize = pop.size();
//   double out1, out2;
//   map < long, vector < int > * > :: iterator mIter;
  
//   for(sIter = PP.begin(); sIter != PP.end(); sIter++)
//     {
//       mIter = (distincts.insert(map < long, vector < int > * > :: value_type((*sIter)->id(), new vector < int >))).first;
      
//       for(int i = 0; i < popSize; i++)
// 	{
// 	  for(int j = 0; j < popSize; j++)
// 	    {
// 	      if(pop[i]->action() == pop[j]->action()) // consider only distinctions between learners of the same class
// 		{
// 		  out1 = (*sIter)->getOutcome(pop[i]->id());
// 		  out2 = (*sIter)->getOutcome(pop[j]->id());
		  
// 		  //	      if(isEqual(out1, out2))
// 		  if(isEqual(out1, out2, 0.1))
// 		    mIter->second->push_back(0);
// 		  else if(out1 > out2)
// 		    mIter->second->push_back(1);
// 		  else
// 		    mIter->second->push_back(0);
// 		}
// 	    }
// 	}
//     }
// }

// bool dominated(vector < double > &is, vector < double > &by)
// {
//   bool equal = true;
  
//   if(is.size() != by.size())
//     die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
//   for(int i = 0; i < is.size(); i++)
//     {
//       if(isEqual(is[i], by[i]) == true) continue;
      
//       equal = false;
      
//       if(is[i] > by[i]) return false;
//     }
  
//   return equal == false;
// }

// bool dominated(vector < double > &is, vector < double > &by, double e)
// {
//   bool equal = true;
  
//   if(is.size() != by.size())
//     die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
//   for(int i = 0; i < is.size(); i++)
//     {
//       if(isEqual(is[i], by[i], e) == true) continue;
      
//       equal = false;
      
//       if(is[i] > by[i]) return false;
//     }
  
//   return equal == false;
// }

// bool dominated(vector < int > &is, vector < int > &by)
// {
//   bool equal = true;
  
//   if(is.size() != by.size())
//     die(__FILE__, __FUNCTION__, __LINE__, "outcome vectors do not match");
  
//   for(int i = 0; i < is.size(); i++)
//     {
//       if(is[i] == by[i]) continue;
      
//       equal = false;

//       if(is[i] > by[i]) return false;
//     }
  
//   return equal == false;
// }

// void auction(Pop &pop, DataSet &data, Point &point)
// {
//   if(pop._agents.size() < 2)
//     die(__FILE__, __FUNCTION__, __LINE__, "not enough agents to auction");

//   double *instance = data.feature(point.id());
//   int outcome = data.category(point.id());
  
//   double due; // what winner must pay out
//   double reward; // what winner receives
  
//   double *REG = (double *) alloca(REGISTERS * sizeof(double));
  
//   Agent *winner = *(pop._agents.begin());
//   winner->_dummy = winner->bid(instance, REG);

//   vector < Agent * >::iterator iter = pop._agents.begin();
  
//   for(iter++; iter != pop._agents.end(); iter++)
//     {
//       (*iter)->_dummy = (*iter)->bid(instance, REG);
      
//       if((*iter)->outbids(*winner))
// 	winner = *iter;
//     }
  
//   // at this point _dummy is the bid of each agent and winner points to the highest bidder
//   // the outcomes should be cleared for the point
  
//   setOutcome(point, pop, outcome, winner);
  
//   due = winner->_dummy;
  
//   if(outcome == winner->_action)
//     reward = 1;
//   else
//     reward = 0;
  
//   winner->_W += (reward - due);
//   winner->recordWin(due, reward);
  
// #ifdef CCM_DEBUG
//   cout << DEBUG << " PAT " << point.id() << " CAT " << outcome << " WINNER " << *winner << " BID " << winner->_dummy << " PAY " << due << endl;
// #endif
// }

// void setOutcome(Point &point, Pop &pop, int outcome, Agent *winner)
// {  
//   for(int i = 0; i < pop.size(); i++) // THIS CAN BE SPED UP IF REWARD IS EITHER 1 0R 0
//     if(pop[i]->action() == outcome)
//       point.setOutcome(pop[i]->id(), pop[i]->_dummy); // 1 - (1 - pop[i]->_dummy)
//     else
//       point.setOutcome(pop[i]->id(), 1.0 - pop[i]->_dummy); // 1 - (0 + pop[i]->_dummy)

			
// //   Agent *aPtr = NULL; // runner up
  
// // #ifdef CCM_DEBUG
// //   cout << DEBUG;
// // #endif
  
// //   for(int i = 0; i < pop.size(); i++) // consider each agent
// //     {
// // #ifdef CCM_DEBUG
// //       cout << " " << *pop[i] << ">>" << pop[i]->_dummy;
// // #endif 

// //       point.setOutcome(pop[i]->id(), 1);
      
// //       if(pop[i]->action() != winner->action()) // class differet from winner
// // 	{
// // 	  if((aPtr == NULL) || (pop[i]->outbids(*aPtr) == true))
// // 	    aPtr = pop[i];
// // 	}
// //     }

// // #ifdef CCM_DEBUG
// //   cout << endl;
// //   if(aPtr != NULL)
// //     cout << DEBUG << " >>APTR" << *aPtr << " BID " << aPtr->_dummy << endl;
// //   else
// //     cout << DEBUG << " >>APTRNULL" << endl;
// // #endif
    
// //   for(int i = 0; i < pop.size(); i++)
// //     {
// //       if((pop[i]->action() == winner->action()) && ((aPtr == NULL) || (pop[i]->outbids(*aPtr))))
// // 	{
// // 	  if(pop[i]->action() == outcome) 
// // 	    point.setOutcome(pop[i]->id(), 2);
// // 	  else
// // 	    point.setOutcome(pop[i]->id(), 0);
// // 	}
// //     }
// }

// void selectPop(Pop &pop, map < long, vector < double > * > &outcomes,  map < int, int > &aCount, stats &s)
// {
//   vector < Agent * > parent;
//   vector < Agent * > child;
  
//   for(int i = 0; i < pop.size(); i++)
//     {
//       if(pop[i]->old() == true)
// 	parent.push_back(pop[i]);
//       else
// 	child.push_back(pop[i]);
//     }
  
//   pop.clear();
  
//   random_shuffle(child.begin(), child.end());
  
//   int pSize = parent.size();
//   int cSize = child.size();
  
//   bool dom;
//   int p, c;
  
//   for(c = 0; c < cSize; c++)
//     {
//       random_shuffle(parent.begin(), parent.end());
      
//       for(p = 0; p < pSize; p++)
// 	{
// 	  //	  if(parent[p]->old() == false || (aCount[parent[p]->action()] < 2 && parent[p]->action() != child[c]->action())) continue;
// 	  if(parent[p]->old() == false || parent[p]->action() != child[c]->action()) continue;
	  
// 	  //	  dom = dominated(*outcomes[parent[p]->id()], *outcomes[child[c]->id()]);
// 	  dom = dominated(*outcomes[parent[p]->id()], *outcomes[child[c]->id()], 0.1);
	  
// 	  if(dom == true)
// 	    {
// 	      parent[p]->old(false);
// 	      child[c]->old(true);
	      
// 	      aCount[parent[p]->action()]--;
// 	      aCount[child[c]->action()]++;

// #ifdef CCM_DEBUG
// 	      cout << DEBUG << " " << *child[c] << " LREPLACES " << *parent[p] << endl;
// #endif
// 	      s.learnerCount++;
	      
// 	      break;
// 	    }
// 	}
//     }
  
//   for(int i = 0; i < pSize; i++)
//     if(parent[i]->old() == true) pop.pushBack(parent[i]);
//     else delete parent[i];
  
//   for(int i = 0; i < cSize; i++)
//     if(child[i]->old() == true) pop.pushBack(child[i]);
//     else delete child[i];
// }

// void selectPoints(set < Point * > &PP, map < long, vector < int > * > &distincts, map < int, int > &pCount, stats &s)
// {
//   vector < Point * > parent;
//   vector < Point * > child;
  
//   set < Point * > ::iterator pIter;
  
//   for(pIter = PP.begin(); pIter != PP.end(); pIter++)
//     {
//       if((*pIter)->old() == true)
// 	parent.push_back(*pIter);
//       else
// 	child.push_back(*pIter);
//     }
  
//   PP.clear();
  
//   random_shuffle(child.begin(), child.end());
  
//   int pSize = parent.size();
//   int cSize = child.size();
  
//   bool dom;
//   int c, p;
  
//   for(c = 0; c < cSize; c++)
//     {
//       random_shuffle(parent.begin(), parent.end());
      
//       for(p = 0; p < pSize; p++)
// 	{
// 	  //	  if(parent[p]->old() == false || (pCount[parent[p]->category()] < 2 && parent[p]->category() != child[c]->category())) continue;
// 	  if(parent[p]->old() == false || parent[p]->category() != child[c]->category()) continue;
	  
// 	  dom = dominated(*distincts[parent[p]->id()], *distincts[child[c]->id()]);
	  
// 	  if(dom == true)
// 	    {
// 	      parent[p]->old(false);
// 	      child[c]->old(true);

// 	      pCount[parent[p]->category()]--;
// 	      pCount[child[c]->category()]++;
	      
// #ifdef CCM_DEBUG
// 	      cout << DEBUG << " " << child[c]->id() << " PREPLACES " << parent[p]->id() << endl;
// #endif
// 	      s.pointCount++;

// 	      break;
// 	    }
// 	}
//     }

//   for(int i = 0; i < pSize; i++)
//     if(parent[i]->old() == true) PP.insert(parent[i]);
//     else delete parent[i];
  
//   for(int i = 0; i < cSize; i++)
//     if(child[i]->old() == true) PP.insert(child[i]);
//     else delete child[i];

// }

// void reduce(Pop &pop, map < long, vector < double > * > &outcomes)
// {
//   for(int i = 0; i < pop.size(); i++)
//     {
//       pop[i]->active(true);
      
//       for(int j = 0; j < pop.size(); j++)
// 	{
// 	  if(pop[i]->action() != pop[j]->action()) continue;
	  
// 	  if(i != j)
// 	    {
// 	      //	      if(dominated(*outcomes[pop[i]->id()], *outcomes[pop[j]->id()]) == true)
// 	      if(dominated(*outcomes[pop[i]->id()], *outcomes[pop[j]->id()], 0.1) == true)
// 		{
// 		  pop[i]->active(false);
// 		  break;
// 		}
// 	    }
// 	}
//     }
// }





























// void reduce(Pop &pop, map < long, int > &ranks) // active true iff learner in LF
// {
//   for(int i = 0; i < pop._agents.size(); i++)
//     {
//       //      if(find(LF.begin(), LF.end(), pop[i]->id()) == LF.end())
//       if(ranks[pop[i]->id()] == 0)
// 	pop[i]->_active = true;
//       //	pop[i]->_active = false; // not in LF
//       else
// 	pop[i]->_active = false;
//       //	pop[i]->_active = true;
//     }
// }



// void initPoints(int numPoints, 
// 		long maxId, // ids must be from 0 to maxId inclusive
// 		set < Point * > &P, // where to store the new points
// 		set < Point * > &E) // the new points can't be the same as points in here
// {
//   // if points in 0, 1, ..., maxId then there are at most maxId + 1 points
  
//   if(numPoints + E.size() > maxId + 1)
//     die(__FILE__, __FUNCTION__, __LINE__, "cannot generate that many points");
  
//   // free old points

//   set < Point * >::iterator oldIter;

//   for(oldIter = P.begin(); oldIter != P.end(); oldIter++)
//     delete *oldIter;
  
//   P.clear();

//   // generate new set of ids

//   set < long > ids;
//   long newPoint;

//   while(ids.size() < numPoints)
//     {
//       newPoint = (long) (drand48() * (maxId + 1));
      
//       while(find_if(E.begin(), E.end(), PointIdEquals(newPoint)) != E.end())
// 	newPoint = (newPoint + 1) % (maxId + 1);
      
//       ids.insert(newPoint); // only insert if not in set E
//     }

//   // create new points using ids

//   set < long >::iterator newIter;

//   for(newIter = ids.begin(); newIter != ids.end(); newIter++)
//     P.insert(new Point(*newIter));
// }


// void reproduce(Pop &pop, long epoch, Args &args, map < long, vector < double > * > &outcomes, map < long, int > &ranks)
// {
//   if(pop._agents.size() < 4)
//     die(__FILE__, __FUNCTION__, __LINE__, "pop size < 4 (too small)");

//   for(int i = 0; i < pop.size(); i++)
//     pop[i]->rank(ranks[pop[i]->id()]);
  
//   stable_sort(pop._agents.begin(), pop._agents.end(), agentPtrGreaterR());
  
// //   for(int i = 0; i < pop.size(); i++)
// //     {
// //       cout << "YEH" << epoch;
// //       for(int j = 0; j < outcomes[pop[i]->id()]->size(); j++) cout << "\t" << (*outcomes[pop[i]->id()])[j];
// //       cout << "\t rank " << pop[i]->rank() << " " << *pop[i] << endl;
// //     }

//   Agent *remd = removeLearner(pop, outcomes);  

//   //  Agent *parent = removeLearner(pop, outcomes);  
  
//   int currAgt = (int) (drand48() * pop.size()); // parent
//   int numTries = 0;

//   while(pop[currAgt]->action() != remd->action())
//     {
//       currAgt = (currAgt + 1) % pop.size();
//       if(++numTries > pop.size())
// 	die(__FILE__, __FUNCTION__, __LINE__, "cannot find agent of the same class");
//     }
  

//   Agent *newAgt = new Agent(args, epoch, *pop[currAgt]);

// //  Agent * newAgt = new Agent(args, epoch, *parent);

// //  cout << DEBUG << " AGENT " << *newAgt << " SPAWN OF " << *parent << endl;
// //  cout << DEBUG << " AGENT " << *newAgt << " SPAWN OF " << *pop[currAgt] << " REPLACED " << *remd << endl;

//   pop.pushBack(newAgt);
//   delete remd;
  
//   //  delete parent;

// //   //  stable_sort(pop._agents.begin(), pop._agents.end(), agentPtrGreaterCW());

// //   //  double pActionMutate;

// //   //  map < int, int > actionCount;
  
// //   int currAgt;
// //   //  int nextAgt;
  
// //   Agent *newAgt;
  
// // #ifdef CCM_DEBUG  
// //   cout << DEBUG << " SORTED";
// //   for(int i = 0; i < pop._agents.size(); i++)
// //     cout << " " << *pop._agents[i];
// //   cout << endl;
// // #endif

// // //   for(int i = 0; i < actions.size(); i++)
// // //     actionCount.insert(map < int, int > :: value_type (actions[i], 0));
  
// // //   for(int i = 0; i < pop.size(); i++)
// // //     actionCount[pop[i]->action()]++;

// // // #ifdef CCM_DEBUG
// // //   cout << DEBUG << " ACTIONCOUNT";
// // //   for(int i = 0; i < actions.size(); i++)
// // //     cout << " " << actions[i] << "->" << actionCount[actions[i]];
// // //   cout << endl;
// // // #endif

// // //   vector < Agent * > :: iterator begin, end;
// // //   for(begin = pop.begin(); begin != pop.end(); begin = end)
// // //     {
// // //       end = begin;
// // //       while(end != pop.end() && (*begin)->action() == (*end)->action()) end++;
      
// // // #ifdef CCM_DEBUG
// // //       cout << DEBUG << " ACTION " << (*begin)->action() << " FROM " << **begin;
// // //       cout << (end == begin ? **begin : **(end - 1)) << endl;
// // // #endif
// // //     }
  
// //   int begin, end; // begin points to first, end points to one past last
// //   for(begin = 0; begin < pop.size(); begin = end)
// //     {
// //       end = begin + 1;
// //       while(end < pop.size() && (pop[begin]->action() == pop[end]->action())) end++;

// // #ifdef CCM_DEBUG
// //       cout << DEBUG << " ACTION " << pop[begin]->action() << " FROM " << begin << " TO " << end << endl;
// // #endif

// //       if(end - begin < 2)
// // 	die(__FILE__, __FUNCTION__, __LINE__, "each action must be represented twice in the population");

// //       currAgt = begin + (int) (drand48() * (end - begin)); // parent
// //       newAgt = new Agent(args, epoch, *pop[currAgt]);

// // #ifdef CCM_DEBUG
// //       cout << DEBUG << " AGENT " << *newAgt << " SPAWN OF " << *pop[currAgt];
// //       cout << " (" << currAgt << ") REPLACING " << *pop[begin] << endl;
// // #endif

// //       delete pop[begin];
// //       pop._agents.erase(pop.begin() + begin);

// //       pop._agents.insert(pop.begin() + begin, newAgt);
// //       pop[begin]->_W = pop[begin + 1]->_W;      
// //     }
    
// // //   for(int i = 0; i < actions.size(); i++)
// // //     if(actionCount[actions[i]] < 2)
// // //       die(__FILE__, __FUNCTION__, __LINE__, "each action must be represented twice in the population");

// // //   int start = 0;
  
// // //   while(start < pop.size())
// // //     {
// // //       if(actionCount[pop[start]->action()] < 2)
// // // 	die(__FILE__, __FUNCTION__, __LINE__, "not enough agents of given class");
      
// // //       currAgt = start + (int) (drand48() * actionCount[pop[start]->action()]); // parent
// // //       newAgt = new Agent(args, epoch, *pop[currAgt]);
      
// // //       delete pop[start];
// // //       pop._agents.erase(pop._agents.begin() + start);

// // //       pop._agents.insert(pop._agents.begin() + start, new Agent(args, epoch, *pop[currAgt]));
// // //       pop[start]->_W = pop[start + 1]->_W;
      
// // //       start += actionCount[pop[start]->action()]; 

// // // #ifdef CCM_DEBUG
// // //       cout << DEBUG << " " << *pop[currAgt] << " CREATED " << *pop[start] << endl;
// // // #endif
// // //     }
  
// // //   // remove poorest agent
  
// // //   delete pop._agents.back();
// // //   pop._agents.pop_back();

// // //   // create new agent
  
// // //   currAgt = (int) (drand48() * pop._agents.size()); // select parent randomly
  
// // // //   for(int t = 0; t < args.getLong("tSize"); t++)
// // // //     {
// // // //       nextAgt = (int) (drand48() * pop._agents.size());
  
// // // //       if(pop[nextAgt]->_W > pop[currAgt]->_W) currAgt = nextAgt;
// // // //     }
  
// // //   pActionMutate = (double) pop.actionCount(pop[currAgt]->action()) / pop.size();

// // //   newAgt = new Agent(args, epoch, *pop[currAgt], actions, pActionMutate);
// // //   newAgt->_W = pop._agents.back()->_W;

// // //   pop._agents.push_back(newAgt);
  
// // // #ifdef CCM_DEBUG
// // //   cout << DEBUG << " " << *pop[currAgt] << " CREATED " << *newAgt << endl;
// // // #endif
// }

// int fnds(map < long, vector < double > * > &outcomes, map < long, int > &ranks)
// {
//   ranks.clear();
  
//   map < long, int > n;
//   map < long, int > :: iterator nIter;
  
//   map < long, set < long > * > S;
//   map < long, set < long > * > :: iterator SIter;
//   set < long > :: iterator sIter;
//   set < long > *s;

//   map < long, vector < double > * > :: iterator oIter1, oIter2;

//   set < long > LS1, LS2;
//   set < long > *F = &LS1;
//   set < long > *Q = &LS2;
//   set < long > :: iterator FIter;

//   int front = 0; // current front being considered

//   for(oIter1 = outcomes.begin(); oIter1 != outcomes.end(); oIter1++)
//     {
//       SIter = (S.insert(map < long, set < long > * > :: value_type (oIter1->first, new set < long >))).first;
//       nIter = (n.insert(map < long, int > :: value_type (oIter1->first, 0))).first;
      
//       for(oIter2 = outcomes.begin(); oIter2 != outcomes.end(); oIter2++)
// 	{
// 	  if(dominated(*(oIter2->second), *(oIter1->second))) (SIter->second)->insert(oIter2->first);
// 	  else if(dominated(*(oIter1->second), *(oIter2->second))) (nIter->second)++;
// 	}

//       if(nIter->second == 0)
// 	{
// 	  ranks.insert(map < long, int > :: value_type(oIter1->first, front));
// 	  F->insert(oIter1->first);
// 	}
//     }

//   while(F->empty() == false)
//     {
//       Q->clear();
//       front++;
      
//       for(FIter = F->begin(); FIter != F->end(); FIter++)
// 	{
// 	  s = (S.find(*FIter))->second;

// 	  for(sIter = s->begin(); sIter != s->end(); sIter++)
// 	    {
// 	      if(--n[*sIter] == 0)
// 		{
// 		  ranks.insert(map < long, int > :: value_type(*sIter, front));
// 		  Q->insert(*sIter);
// 		}
// 	    }
// 	}
      
//       if(F == &LS1)
// 	{
// 	  F = &LS2;
// 	  Q = &LS1;
// 	}
//       else
// 	{
// 	  F = &LS1;
// 	  Q = &LS2;
// 	}
      
//     }

//   for(SIter = S.begin(); SIter != S.end(); SIter++)
//     delete SIter->second;

//   return front;
// }

//void selectPop(Pop &pop, map < long, vector < int > * > &outcomes,  map < int, int > &aCount)
//{
  
//   vector < Agent * > :: iterator aIter;
  
//   set < long > toRemove;
//   bool dom;
  
//   for(aIter = pop.begin(); aIter != pop.end(); aIter++)
//     {
//       if((*aIter)->old() == false) // a child
//   	{
// 	  dom = dominated(*outcomes[(*aIter)->parentPtr()->id()], *outcomes[(*aIter)->id()]);
	  
// 	  if(dom == false && 
// 	     isEqual(*outcomes[(*aIter)->id()], *outcomes[(*aIter)->parentPtr()->id()]) == true && 
// 	     ((*aIter)->action() == (*aIter)->parentPtr()->action()))
// 	    if(drand48() < 0.5) dom = true;
	  
//   	  if(dom == true) // dominates parent
// 	    {
// 	      toRemove.insert((*aIter)->parentPtr()->id());
// 	      aCount[(*aIter)->action()]++;
// 	      aCount[(*aIter)->parentPtr()->action()]--;
// 	    }
// 	  else
// 	    {
// 	      toRemove.insert((*aIter)->id());
// 	    }
	  
// #ifdef CCM_DEBUG
// 	  if(dom == true)
// 	    cout << DEBUG << " " << **aIter << " VS " << (*aIter)->parentPtr()->id() << " BETTER " << endl;
// 	  else
// 	    cout << DEBUG << " " << **aIter << " VS " << (*aIter)->parentPtr()->id() << " WORSE " << endl;
// #endif
// 	}
//     }
  
//   for(aIter = pop.begin(); aIter != pop.end(); )
//     {
//       if(toRemove.find((*aIter)->id()) != toRemove.end())
// 	{
// 	  delete *aIter;
// 	  aIter = pop.remove(aIter);
// 	}
//       else
// 	{
// 	  (*aIter)->old(true);
// 	  aIter++;
// 	}
//     }
//}


  
//   int TSIZE = 4;
  
//   vector < Point * > parent;
//   vector < Point * > child;
  
//   set < Point * > :: iterator pIter;
  
//   int curr, next;
//   int minDist, dist;
  
//   for(pIter = PP.begin(); pIter != PP.end(); pIter++)
//     {
//       if((*pIter)->old() == true)
// 	parent.push_back(*pIter);
//       else
// 	child.push_back(*pIter);
//     }
  
//   PP.clear();
  
//   int parentSize = parent.size();
//   int childSize = child.size();
  
//   bool dom;
  
//   if(parentSize != childSize)
//     die(__FILE__, __FUNCTION__, __LINE__, "parentSize != childSize");
  
// #ifdef CCM_DEBUG
//   cout << DEBUG << " PCSIZE " << parentSize << endl;
// #endif
  
//   for(int c = 0; c < childSize; c++)
//     {
//       curr  = (int) (drand48() * parentSize);
//       while(parent[curr]->old() == false) curr = (curr + 1) % parentSize;
//       minDist = hammingDist(*distincts[child[c]->id()], *distincts[parent[curr]->id()]);
      
//       for(int t = 0; t < TSIZE; t++)
// 	{
// 	  next  = (int) (drand48() * parentSize);
// 	  while(parent[next]->old() == false) next = (next + 1) % parentSize;
// 	  dist = hammingDist(*distincts[child[c]->id()], *distincts[parent[next]->id()]);
	  
// 	  if(dist < minDist)
// 	    {
// 	      curr = next;
// 	      minDist = dist;
// 	    }
// 	}
      
//       dom = dominated(*distincts[parent[curr]->id()], *distincts[child[c]->id()]);
      
//       if(dom == true)
// 	{
// 	  parent[curr]->old(false);
// 	  child[c]->old(true);
// 	}
      
// #ifdef CCM_DEBUG
//       if(dom == true)
// 	cout << DEBUG << " " << child[c]->id() << " REPLACES " << parent[curr]->id() << endl;
//       else
// 	  cout << DEBUG << " " << child[c]->id() << " YIELDS " << parent[curr]->id() << endl;
// #endif	
//     }
  
//   for(int i = 0; i < parentSize; i++)
//     if(parent[i]->old() == true) PP.insert(parent[i]);
//     else delete parent[i];
  
//   for(int i = 0; i < childSize; i++)
//     if(child[i]->old() == true) PP.insert(child[i]);
//     else delete child[i];



