#ifndef AGENT_H
#define AGENT_H

#include <vector>
#include <bitset>
#include <cmath>
#include <sstream>
#include <cfloat>

#include "args.hpp"
#include "dataset.hpp"
#include "misc.hpp"

using namespace std;

#define REGISTERS 8
#define INPUTS 64

typedef bitset < 6 + 3 + 3 + 1 > instruction;

class Agent
{
  
public:
  
  Agent(Args &, long, long); // Create agent from scratch.
  Agent(Args &, long, Agent &, vector < int > &, double); // Create agent from another.
  
  ~Agent();
  
  friend ostream & operator<<(ostream &, const Agent &); // Agent members.
  
  double bid(double *, double *);
  double run(vector < instruction * > &, double *, double *);
  
  inline int size() { return _bid.size(); }
  inline long id() { return _id; }
  inline int action() { return _action; }
  inline long parentId() { return _parentId; }
  inline Agent * parentPtr() { return _parentPtr; }

  inline bool active() { return _active; }
  inline void active(bool a) { _active = a; }

  inline int rank() { return _rank; }
  inline void rank(int r) { _rank = r; }

  inline bool old() { return _old; }
  inline void old(bool o) { _old = o; }
  
  inline bool outbids(Agent &comp) { return isEqual(_dummy, comp._dummy) ? winsTie(comp) : _dummy > comp._dummy; }
  bool winsTie(Agent &);
  
  void recordWin(double, double);

  long _numWon; // number of auctions won
  double _sumDue; // total payments made over the _numWon auctions
  double _sumReward; // total reward received over the _numWon auctions

  long _numProfitWon; // number of times a profit was made
  double _sumProfitDue; // total payments made over the _numProfitWon auctions
  double _sumProfitReward; // total reward received over the _numProfitWon auctions

  double _profitRate; // fraction of auctions won that are also profitable
  
  static unsigned long _count; // Next id to use.
  
  int _dim;
  unsigned long _id;
  int _action;
  double _W;
  long _birthEpoch;
  long _parentId;
  Agent *_parentPtr;
  
  vector < instruction * > _bid; // Bid program.
  
  // Instruction definitions.
  
  // Modes.
  
  static const instruction _mode0;
  static const instruction _mode1;
  
  // Masks. Apply before shifting.
  
  static const instruction _modeMask;
  static const instruction _opMask;
  static const instruction _dstMask;
  static const instruction _srcMask;
  
  // Binary operations.
  
  static const instruction _opSum;
  static const instruction _opDiff;
  static const instruction _opProd;
  static const instruction _opDiv;
  static const instruction _opMod;
  static const instruction _opCos;
  static const instruction _opExp;
  static const instruction _opLog;

  // Shift amounts.

  static const short _modeShift;
  static const short _opShift;
  static const short _dstShift;
  static const short _srcShift;

  // Bookkeeping stuff, not part of CCM.
  
  double _dummy; //
  bool _active; //
  int _rank;

  bool _old;
};

struct agentPtrGreaterW:public binary_function < Agent *, Agent *, bool >
{
  bool operator()(Agent *lhs, Agent *rhs) { return lhs->_W > rhs->_W; }
};

struct agentPtrGreaterR:public binary_function < Agent *, Agent *, bool >
{
  bool operator()(Agent *lhs, Agent *rhs) { return lhs->_rank < rhs->_rank; }
};


struct agentPtrGreaterCW:public binary_function < Agent *, Agent *, bool >
{
  bool operator()(Agent *lhs, Agent *rhs) { return lhs->_action == rhs->_action? lhs->_W < rhs->_W : lhs->_action < rhs->_action; }
};


#endif
